//
//  kmpg
//
//  Copyright (C) 1997 Bjarni Ivarsson
//  email: Bjarni.Ivarsson@swipnet.se
//
//  This program is free software; you can redistribute it and/or modify
//  it under the terms of the GNU General Public License as published by
//  the Free Software Foundation; either version 2 of the License, or
//  (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//  You should have received a copy of the GNU General Public License
//  along with this program; if not, write to the Free Software
//  Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.
//

#include "ledpanel.h"

#include <iostream.h>
#include <stdlib.h>
#include <qpainter.h>
#include <qregexp.h>

#include <kapp.h>
#include <kiconloader.h>
#include <yafBridge/frontendDecoder.h>
#include "playlistmgr.h"
#include "led_pixmaps.h"

#define ADD_BUTTON( btn, x, y, w, h, object, slot, icon ) \
   btn = new QPushButton( "", _controlPanel ); \
   btn->setGeometry( x, y, w, h ); \
   btn->setPixmap( loader->loadIcon(icon) ); \
   connect( btn, SIGNAL(clicked()), object, SLOT(slot) )

#define ENABLE_BUTTONS( e1, e2, e3, e4, e5, e6, e7 ) \
   _pbPrev->setEnabled( e1 ); \
   _pbBwd->setEnabled( e2 ); \
   _pbPause->setEnabled( e3 ); \
   _pbPlay->setEnabled( e4 ); \
   _pbStop->setEnabled( e5 ); \
   _pbFwd->setEnabled( e6 ); \
   _pbNext->setEnabled( e7 )

// Toggle the pause and play buttons
#define TOGGLE_PP( e1, e2 ) \
   _pbPause->setOn( e1 ); \
   _pbPlay->setOn( e2 )



LedPanel::LedPanel( QWidget *parent, PlaylistMgr *mgr, FrontendDecoder *decoder ):
   QWidget( parent )
{
   int height;
   _mgr = mgr;
   _decoder = decoder;
   connect( mgr, SIGNAL(sigChanged( bool )), SLOT(setListStatus( bool )) );
   loadPixmaps();
   pm_back.resize( 300, 200 );
   pm_back.fill(QColor(0,0,0));
   bitBlt(&pm_panel, pm_panel.width()-43 , 0, &pm_dots );
   height = pm_panel.height();

   _statusLine = new StatusLine( this, pm_panel.width() );
   _statusLine->move( 0, height );
   connect( _statusLine, SIGNAL(statusLineClicked(int)), SLOT(StatusClicked(int)) );
   height += _statusLine->height();

   _infoLine = new InfoLine( this, mgr, pm_panel.width() );
   _infoLine->move( 0, height );
   height += _infoLine->height();


   _trackCalendar = new TrackCalendar( this, mgr );
   _trackCalendar->move( 0, height );
   connect( _trackCalendar, SIGNAL(sigSizeChanged()), SLOT(doSizeChanged()) );
   _expanded = TRUE; // start with the trackcalendar showing
   height += _trackCalendar->height();

   // create slider
   _sliderPanel = new QWidget( this );
   _sliderPanel->setGeometry( 0, height, 207, 20 );
   _slider = new KSlider( KSlider::Horizontal, _sliderPanel );
   _slider->setGeometry( 2, 1, 202, 20 );
   _slider->setRange( 0, 300 );
   _slider->setSteps( 15, 60 );
   connect( this, SIGNAL(seek(int)), this, SLOT(seekTo( int )) );
   connect( _slider, SIGNAL(sliderPressed()), SLOT(disableUpdate()) );
   connect( _slider, SIGNAL(sliderReleased()), SLOT(enableUpdate()) );
   connect( _slider, SIGNAL(sliderMoved(int)), SLOT(sliderFrame(int)) );
   connect( decoder, SIGNAL(time(float)), SLOT(decoderTime(float)) );
   height += _sliderPanel->height();

   // create controls
   _controlPanel = new QWidget( this );
   _controlPanel->setGeometry( 0, height, 207, 26 );
   KIconLoader *loader = kapp->getIconLoader();
   ADD_BUTTON( _pbPrev, 5, 2, 28, 22, mgr, playPrev(),     "kmpg_prev.xpm" );
   ADD_BUTTON( _pbBwd, 33, 2, 28, 22, this, rewind(),      "kmpg_rew.xpm" );
   ADD_BUTTON( _pbPause, 61, 2, 28, 22, this, dopause(),  "kmpg_pause_off.xpm" );
   ADD_BUTTON( _pbPlay, 89, 2, 28, 22, this, doplay(),     "kmpg_play_off.xpm" );
   ADD_BUTTON( _pbStop, 117, 2, 28, 22, decoder, stop(),   "kmpg_stop_off.xpm" );
   ADD_BUTTON( _pbFwd, 145, 2, 28, 22, this, forward(),    "kmpg_fwd.xpm" );
   ADD_BUTTON( _pbNext, 173, 2, 28, 22, mgr, playNext(),   "kmpg_next.xpm" );   
   _pbPause->setToggleButton( TRUE );
   _pbPlay->setToggleButton( TRUE );
   ENABLE_BUTTONS( FALSE, FALSE, FALSE, FALSE, FALSE, FALSE, FALSE );
   TOGGLE_PP( FALSE, FALSE );
   connect( decoder, SIGNAL(playing()), SLOT(play()) );
   connect( decoder, SIGNAL(paused()), SLOT(pause()) );
   height += _controlPanel->height();
   
   _showTrackCal = TRUE;
   _showSlider = TRUE;
   _showControls = TRUE;
   _update = TRUE;
   
   pm_back.resize( pm_panel.width(), 200 );
   bitBlt(&pm_back, 0, 0, &pm_panel );
   setFixedSize(pm_panel.width(), height);
   adjustSize();
   setBackgroundPixmap( pm_back );
   doSizeChanged();
   connect( this, SIGNAL(timeDisplayChanged(bool)), SLOT(ledPanelTimeMode(bool)) );
   for( int i=0; i<10; i++ ) value_digits[i] = -2;

   switch(_mgr->getPlayMode())
   {
      case PlaylistMgr::Sequential:
	 setField(2, "sequential");
	 break;
      case PlaylistMgr::Random:
	 setField(2, "random");
	 break;
      default:
	 break;
   }

   switch(_mgr->getRepeatMode())
   {
      case PlaylistMgr::Once:
         setField(3, "rep. off");
         break;
      case PlaylistMgr::All:
         setField(3, "rep. list");
         break;
      case PlaylistMgr::One:
         setField(3, "rep. track");
         break;
      default:
         break;
   }
}

LedPanel::~LedPanel()
{
   delete _infoLine;
}

void LedPanel::mouseDoubleClickEvent( QMouseEvent * )
{
   _expanded = !_expanded;
   doSizeChanged();
}

void LedPanel::mousePressEvent( QMouseEvent * e )
{
   emit(timeDisplayChanged(TRUE));
   cerr << "timeModeChanged" << endl;
}

void LedPanel::loadPixmaps()
{
   pm_digits[0] = QPixmap(digit_0);
   pm_digits[1] = QPixmap(digit_1);
   pm_digits[2] = QPixmap(digit_2);
   pm_digits[3] = QPixmap(digit_3);
   pm_digits[4] = QPixmap(digit_4);
   pm_digits[5] = QPixmap(digit_5);
   pm_digits[6] = QPixmap(digit_6);
   pm_digits[7] = QPixmap(digit_7);
   pm_digits[8] = QPixmap(digit_8);
   pm_digits[9] = QPixmap(digit_9);
   pm_panel     = QPixmap(panel);
   pm_dots      = QPixmap(points);
   pm_blank     = QPixmap(digit_blank);
   pm_pause     = QPixmap(digit_pause);
   pm_play      = QPixmap(digit_play);
   pm_minus     = QPixmap(digit_minus);
}

void LedPanel::updateDigit( int _digit, int _value )
{
   QPixmap pm;

   if( value_digits[_digit] == _value ) return;
   
   value_digits[_digit] = _value;
   if( _digit == 0 )
      switch( _value )
      {
	 case 1:
	    pm = pm_pause;
	    break;
	 case 2:
	    pm = pm_play;
	    break;
	 default:
	    pm = pm_blank;
	    break;
      }
   else
      if( _value>-1 && _value<10 )
	 pm = pm_digits[ _value ];
      else if (_value > 9)
         pm = pm_minus;
      else
	 pm = pm_blank;

   bitBlt(&pm_back, 1+18*_digit+(_digit>8?9:0), 2, &pm );
   setBackgroundPixmap( pm_back );
}

void LedPanel::play()
{
   setListStatus(TRUE);
   updateDigit( 0, 2 );
   ENABLE_BUTTONS( TRUE, TRUE, TRUE, FALSE, TRUE, TRUE, TRUE );
   TOGGLE_PP( FALSE, TRUE );
}

// hack to work with the statusdevice
void LedPanel::dopause() {
  _decoder->pause();
  _decoder->setstatus(FrontendDecoder::Paused);
}
  


void LedPanel::pause()
{
   updateDigit( 0, 1 );
   ENABLE_BUTTONS( FALSE, FALSE, TRUE, FALSE, FALSE, FALSE, FALSE );
   TOGGLE_PP( TRUE, FALSE );
}

void LedPanel::stop()
{
   updateDigit( 0, 0 );
   TOGGLE_PP( FALSE, FALSE );
   ENABLE_BUTTONS( FALSE, FALSE, FALSE, TRUE, FALSE, FALSE, FALSE );
}

void LedPanel::doSizeChanged()
{
   int y = pm_panel.height()+_infoLine->height()+_statusLine->height();

   if ( _showTrackCal )
      y += _trackCalendar->height();

   if ( _showSlider )
   {
      _sliderPanel->move( 0, y );
      y += _sliderPanel->height();
   }
   else
      _sliderPanel->move( 0, -100 );

   if ( _showControls )
   {
      _controlPanel->move( 0, y );
      y += _controlPanel->height();
   }
   else
      _controlPanel->move( 0, -100 ); 

   setFixedSize(pm_panel.width(), y);
   emit(sizeChanged());
}

void LedPanel::setLeds(int _track, int _secs )
{
   int digits, digit, mins, secs;
   bool negative = FALSE;

   mins = abs(_secs/60);
   secs = abs(_secs%60);
   if ( _secs < 0 ) negative = TRUE;
//   cerr << " seconds: " << _secs << endl;
   if ( mins > 99999 ) mins = mins % 100000;

   if ( _track > 0 )
   {
      digits = 5;
      setTrack(_track);
   }
   else
      digits = 0;

   for ( digit=8; digit > digits ; digit-- )
   {
      if ( mins != 0 )
         updateDigit( digit , mins%10 );
      else if ( digit == 8)
         updateDigit( digit , 0 );
      else if ( negative )
      {
         updateDigit( digit , 10 );
	 negative = FALSE;
      }
      else
	 updateDigit( digit ,-1 );
      mins = mins/10;
   }
   updateDigit( 9, secs/10 );
   updateDigit(10, secs%10 );
}

void LedPanel::setTrack( int track )
{
   if ( track > 9999 ) track = track % 10000;

   track > 999 ? updateDigit( 1, track/1000)   : updateDigit( 1, -1);
   track > 99  ? updateDigit( 2, track/100%10) : updateDigit( 2, -1);
   track > 9   ? updateDigit( 3, track/10%10)  : updateDigit( 3, -1);
   updateDigit( 4, track%10 );
   updateDigit( 5, -1 );
}

void LedPanel::setInfo( Fileinfo &finfo )
{
   _finfo = finfo;
   _infoLine->setInfo( _finfo );
   _slider->setRange( 0, _finfo.getMinNumberOfFrames());
   int st1=(int)(15000/_finfo.getMsPerFrame()), st2=(int)(60000/_finfo.getMsPerFrame());
   float len=_finfo.getMinNumberOfFrames()*_finfo.getMsPerFrame();
   if ( len>=360000 )
      st1 = st1*2;
   while ( len>=480000 )
   {
      st1 *= 2;
      st2 = st1;
      len /= 2;
   }
   _slider->setValue( 0 );
   _slider->setSteps( st1, st2 );
   total_sec = (int)(_finfo.getMinNumberOfFrames()*_finfo.getMsPerFrame()/1000);
   return;
}

void LedPanel::frame(int _frame)
{
   int sec, ms_per_frame;

   ms_per_frame = (int)_finfo.getMsPerFrame();
   
   //current_frame = _frame;
   sec=(int)(current_frame*ms_per_frame/1000);
   if (sec!=current_sec)
   {
      current_sec=sec;
      if (!_update)
         ledSetTime(track, 0, 0, total_sec, current_sec );
      else 
      {
        _slider->setValue(_frame);
        ledSetTime(timedisplay, _mgr->getListTime(), _mgr->getPlayTime(), total_sec, current_sec );
      }
   }
}

void LedPanel::forward() 
{
   int ms_per_frame;

   ms_per_frame = (int)_finfo.getMsPerFrame();
   seekTo(current_frame + (ms_per_frame * 15) );
   cerr << "seeking: " << current_frame << " -> " << current_frame + (ms_per_frame * 15) << endl;

}

void LedPanel::rewind() 
{
   int ms_per_frame;

   ms_per_frame = (int)_finfo.getMsPerFrame();
   seekTo(current_frame - (ms_per_frame * 15) );
   cerr << "seeking: " << current_frame << " -> " << current_frame - (ms_per_frame * 15) << endl;

}

void LedPanel::ledSetTime(TimeMode mode, int list_total, int list_played, int track_total, int track_played) 
{
   int sec, track_no;
   QString string;

   if ( _mgr->isList() )
   {
      track_no = _mgr->getTrack();
      switch(mode) {
         case track:
            sec = track_played;
	    string = "track";
            break;
         case trackrem:
            sec = track_played - track_total;
	    string = "track";
	    break;
	 case total:
	    sec = list_played + track_played;
	    string = "list";
	    track_no = 0;
	    break;
	 case totalrem:
	    sec = list_played + track_played - list_total;
	    string = "list";
	    track_no = 0;
	    break;
         default:
	    sec = 5192;
	    track_no = 0;
            break;
      }
   }
   else 
   {
      track_no = -1;
      if ( mode == total ||  mode == track)
      {
         sec = track_played;
	 string = "elapsed";
      }
      else
      {
         sec = track_played - track_total;
	 string = "remaining";
      }
      setField(0,"-/-");
   }

//      cerr << "timedisplay(" << timedisplay << "): " << sec << "	lt: " << list_total << " lp: " << list_played << "	tt: " << track_total << " tp: " << track_played  << endl;
      setLeds( track_no, sec );
      setField(1 , string);
}

void LedPanel::ledPanelTimeMode(bool change)
{
   int sec;
   if (change)
      switch(timedisplay) {
      case track:
         timedisplay = trackrem;
         break;
      case trackrem:
         timedisplay = total;
         break;
      case total:
         timedisplay = totalrem;
         break;
      case totalrem:
         timedisplay = track;
         break;
      default:
         timedisplay = track;
         sec = 5192;
         break;
      }
   if ( _mgr->isList() )
      ledSetTime(timedisplay, _mgr->getListTime(), _mgr->getPlayTime(), total_sec, current_sec );
   else
      ledSetTime(timedisplay, 0, 0, total_sec, current_sec );
}

void LedPanel::disableUpdate() 
{
   _update = FALSE;
}

void LedPanel::enableUpdate()
{
  int nPos;          
  nPos=_slider->value();
  current_frame=nPos;
  seekTo(nPos);        
  _update = TRUE;
}


void LedPanel::sliderFrame(int i)
{
   if ( !_update )
      frame(i);
}

void LedPanel::paintEvent( QPaintEvent* )
{
}

void LedPanel::styleChange( GUIStyle )
{
}

void LedPanel::showTrackCal( bool s )
{
   _showTrackCal = s;
   doSizeChanged();
}

void LedPanel::showSlider( bool s )
{
   _showSlider = s;
   doSizeChanged();
}

void LedPanel::showControls( bool s )
{
   _showControls = s;
   doSizeChanged();
}

void LedPanel::StatusClicked(int field)
{
   QString string;
   PlaylistMgr::PlayMode pmode;
   PlaylistMgr::RepeatMode rmode;

   if ( field == 1 )
      emit(timeDisplayChanged(TRUE));
   else if ( field == 2 )
      switch(_mgr->getPlayMode())
      {
         case PlaylistMgr::Sequential:
	    _mgr->setPlayMode(PlaylistMgr::Random);
	    setField(2, "random");
	    break;
	 case PlaylistMgr::Random:
	    _mgr->setPlayMode(PlaylistMgr::Sequential);
	    setField(2, "sequential");
	    break;
	 default:
	    break;
      }
   else if ( field == 3 )
      switch(_mgr->getRepeatMode())
      {
         case PlaylistMgr::Once:
	    _mgr->setRepeatMode(PlaylistMgr::All);
	    setField(3, "rep. list");
	    break;
	 case PlaylistMgr::All:
	    _mgr->setRepeatMode(PlaylistMgr::One);
	    setField(3, "rep. track");
	    break;
	 case PlaylistMgr::One:
	    _mgr->setRepeatMode(PlaylistMgr::Once);
	    setField(3, "rep. off");
	    break;
	 default:
	    break;
      }
}

void LedPanel::doplay()
{
   if ( _mgr->areAllPlayed() )
      _mgr->setAllPlayed(FALSE);
   _mgr->playNext();
}

void LedPanel::setField ( int i , QString string )
{
   _statusLine->setField(i , string);
}

void LedPanel::setListStatus ( bool i)
{
   QString string = "", str = "";

   string.setNum(_mgr->getPlayed());
   string += "/";
   str.setNum(_mgr->getTracks());
   string += str;
   setField( 0, string );
}



void LedPanel::seekTo(int frame) { 
  float ms_per_frame =_finfo.getMsPerFrame();
  _decoder->seek(((float)frame*ms_per_frame)/1000.0);  
} 



void LedPanel::decoderTime( float i)
{
  float msPerFrame=_finfo.getMsPerFrame(); 
  current_frame = (int)(i*1000.0/msPerFrame);
  if (_update)
     frame(current_frame);                                                     
}       
